var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
import _ from 'lodash';
import { Operation } from './types';
import { getTypeOfObj, isLeafNode } from './utils';
var _intersection = _.intersection;
var _difference = _.difference;
var _keyBy = _.keyBy;
var Changeset = /** @class */ (function () {
    function Changeset() {
    }
    // embededObjKeys: define keys for arrays
    Changeset.prototype.diff = function (oldObj, newObj, embededObjKeys) {
        return compare(oldObj, newObj, [], embededObjKeys, []);
    };
    ;
    Changeset.prototype.applyChanges = function (obj, changesets) {
        var results = []; // result of each operation, not usefull
        changesets.forEach(function (change) {
            if (isLeafNode(change)) {
                results.push(applyLeafChange(obj, change, change.embededKey));
            }
            else {
                results.push(applyBranchChange(obj[change.key], change));
            }
        });
        return results;
    };
    ;
    Changeset.prototype.revertChanges = function (obj, changesets) {
        var results = []; // result of each operation, not usefull
        changesets.reverse().forEach(function (change) {
            if (!change.changes) {
                results.push(revertLeafChange(obj, change));
            }
            else {
                results.push(revertBranchChange(obj[change.key], change));
            }
        });
        return results;
    };
    ;
    return Changeset;
}());
var changeset = new Changeset();
export default changeset;
var getKey = function (path) {
    var ref = path[path.length - 1];
    return ref !== null ? ref : '$root';
};
var compare = function (oldObj, newObj, path, embededObjKeys, keyPath) {
    var changes = [];
    var typeOfOldObj = getTypeOfObj(oldObj);
    var typeOfNewObj = getTypeOfObj(newObj);
    var diffs;
    if (typeOfOldObj !== typeOfNewObj) {
        changes.push({
            type: Operation.REMOVE,
            key: getKey(path),
            value: oldObj
        });
        changes.push({
            type: Operation.ADD,
            key: getKey(path),
            value: newObj
        });
        return changes;
    }
    switch (typeOfOldObj) {
        case 'Date':
            changes = __spreadArray(__spreadArray([], changes, true), comparePrimitives(oldObj.getTime(), newObj.getTime(), path), true);
            break;
        case 'Object':
            diffs = compareObject(oldObj, newObj, path, embededObjKeys, keyPath);
            if (diffs.length) {
                if (path.length) {
                    changes = __spreadArray(__spreadArray([], changes, true), [{
                            type: Operation.UPDATE,
                            key: getKey(path),
                            changes: diffs
                        }], false);
                }
                else {
                    changes = __spreadArray(__spreadArray([], changes, true), diffs, true);
                }
            }
            break;
        case 'Array':
            changes = __spreadArray(__spreadArray([], changes, true), compareArray(oldObj, newObj, path, embededObjKeys, keyPath), true);
            break;
        case 'Function':
            break;
        default:
            changes = __spreadArray(__spreadArray([], changes, true), comparePrimitives(oldObj, newObj, path), true);
    }
    return changes;
};
var compareObject = function (oldObj, newObj, path, embededObjKeys, keyPath, skipPath) {
    if (skipPath == null) {
        skipPath = false;
    }
    var changes = [];
    var oldObjKeys = Object.keys(oldObj);
    var newObjKeys = Object.keys(newObj);
    var intersectionKeys = _intersection(oldObjKeys, newObjKeys);
    for (var i = 0, len = intersectionKeys.length; i < len; i++) {
        var k = intersectionKeys[i];
        var newPath = path.concat([k]);
        var newKeyPath = skipPath ? keyPath : keyPath.concat([k]);
        var diffs = compare(oldObj[k], newObj[k], newPath, embededObjKeys, newKeyPath);
        if (diffs.length) {
            changes = changes.concat(diffs);
        }
    }
    var addedKeys = _difference(newObjKeys, oldObjKeys);
    for (var i = 0, len = addedKeys.length; i < len; i++) {
        var k = addedKeys[i];
        var newPath = path.concat([k]);
        //const newKeyPath = skipPath ? keyPath : keyPath.concat([k]);
        changes.push({
            type: Operation.ADD,
            key: getKey(newPath),
            value: newObj[k]
        });
    }
    var deletedKeys = _difference(oldObjKeys, newObjKeys);
    for (var i = 0, len = deletedKeys.length; i < len; i++) {
        var k = deletedKeys[i];
        var newPath = path.concat([k]);
        //const newKeyPath = skipPath ? keyPath : keyPath.concat([k]);
        changes.push({
            type: Operation.REMOVE,
            key: getKey(newPath),
            value: oldObj[k]
        });
    }
    return changes;
};
var compareArray = function (oldObj, newObj, path, embededObjKeys, keyPath) {
    var _ref = embededObjKeys != null ? embededObjKeys[keyPath.join('.')] : void 0;
    var uniqKey = _ref != null ? _ref : '$index';
    var indexedOldObj = convertArrayToObj(oldObj, uniqKey);
    var indexedNewObj = convertArrayToObj(newObj, uniqKey);
    var diffs = compareObject(indexedOldObj, indexedNewObj, path, embededObjKeys, keyPath, true);
    if (diffs.length) {
        return [
            {
                type: Operation.UPDATE,
                key: getKey(path),
                embededKey: uniqKey,
                changes: diffs
            }
        ];
    }
    else {
        return [];
    }
};
var convertArrayToObj = function (arr, uniqKey) {
    if (uniqKey !== '$index') {
        return _keyBy(arr, uniqKey);
    }
    else {
        return arr.reduce(function (acc, cur, idx) {
            acc[idx] = cur;
            return acc;
        }, {});
    }
};
var comparePrimitives = function (oldObj, newObj, path) {
    var changes = [];
    if (oldObj !== newObj) {
        changes.push({
            type: Operation.UPDATE,
            key: getKey(path),
            value: newObj,
            oldValue: oldObj
        });
    }
    return changes;
};
var isEmbeddedKey = function (key) {
    return /\$.*=/gi.test(key);
};
var removeKey = function (obj, key, embededKey) {
    var index;
    if (Array.isArray(obj)) {
        if (embededKey !== '$index' || !obj[key]) {
            index = indexOfItemInArray(obj, embededKey, key);
        }
        return obj.splice(index != null ? index : key, 1);
    }
    else {
        return delete obj[key];
    }
};
var indexOfItemInArray = function (arr, key, value) {
    for (var index = 0; index < arr.length; index++) {
        var item = arr[index];
        if (key === '$index') {
            if (item === value) {
                return index;
            }
        }
        else if (item[key] === value) {
            return index;
        }
    }
    return -1;
};
var modifyKeyValue = function (obj, key, value) {
    return obj[key] = value;
};
var addKeyValue = function (obj, key, value) {
    if (Array.isArray(obj)) {
        return obj.push(value);
    }
    else {
        return obj[key] = value;
    }
};
var parseEmbeddedKeyValue = function (key) {
    var uniqKey, value;
    uniqKey = key.substring(1, key.indexOf('='));
    value = key.substring(key.indexOf('=') + 1);
    return {
        uniqKey: uniqKey,
        value: value
    };
};
var applyLeafChange = function (obj, change, embededKey) {
    var type = change.type, key = change.key, value = change.value;
    switch (type) {
        case Operation.ADD:
            return addKeyValue(obj, key, value);
        case Operation.UPDATE:
            return modifyKeyValue(obj, key, value);
        case Operation.REMOVE:
            return removeKey(obj, key, embededKey);
    }
};
var applyArrayChange = function (arr, change) {
    var _a;
    var results = [];
    (_a = change.changes) === null || _a === void 0 ? void 0 : _a.forEach(function (subchange) {
        if ((subchange.value != null) || subchange.type === Operation.REMOVE) {
            results.push(applyLeafChange(arr, subchange, change.embededKey));
        }
        else {
            var element = void 0;
            if (change.embededKey === '$index') {
                element = arr[+subchange.key];
            }
            else {
                element = arr.find(function (el) { return el[change.embededKey] == subchange.key; }); // can only use == to compare 
            }
            results.push(changeset.applyChanges(element, subchange.changes));
        }
    });
    return results;
};
var applyBranchChange = function (obj, change) {
    if (Array.isArray(obj)) {
        return applyArrayChange(obj, change);
    }
    else {
        return changeset.applyChanges(obj, change.changes);
    }
};
var revertLeafChange = function (obj, change, embededKey) {
    var type = change.type, key = change.key, value = change.value, oldValue = change.oldValue;
    switch (type) {
        case Operation.ADD:
            return removeKey(obj, key, embededKey);
        case Operation.UPDATE:
            return modifyKeyValue(obj, key, oldValue);
        case Operation.REMOVE:
            return addKeyValue(obj, key, value);
    }
};
var revertArrayChange = function (arr, change) {
    var _a;
    var results = [];
    (_a = change.changes) === null || _a === void 0 ? void 0 : _a.forEach(function (subchange) {
        if ((subchange.value != null) || subchange.type === Operation.REMOVE) {
            results.push(revertLeafChange(arr, subchange, change.embededKey));
        }
        else {
            var element = void 0;
            if (change.embededKey === '$index') {
                element = arr[+subchange.key];
            }
            else {
                element = arr.find(function (el) { return el[change.embededKey] == subchange.key; });
            }
            results.push(changeset.revertChanges(element, subchange.changes));
        }
    });
    return results;
};
var revertBranchChange = function (obj, change) {
    if (Array.isArray(obj)) {
        return revertArrayChange(obj, change);
    }
    else {
        return changeset.revertChanges(obj, change.changes);
    }
};
//# sourceMappingURL=changesets.js.map